/***************************************************************************
 * jtalk -- talk-like Jabber client
 * (C) Nicolas George - 2006
 * loudmouth.c: interface with the loudmouth library
 ***************************************************************************/

/*
  This file is part of jtalk.

  jtalk is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  jtalk is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with jtalk; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <loudmouth/loudmouth.h>

#include "jtalk.h"

struct jtalk_connection_s {
    const char *server;
    const char *user;
    const char *pass;
    const char *resource;
    const char *jid;
    LmConnection *con;
    LmSSL *tls;
    LmMessageHandler *handler;
    GSList *pending_messages;
    int port;
    int reconnect_delay;
    unsigned connected : 1;
};

LmMessageHandler *handler;

static void jtalk_connection_schedule_connect(jtalk_connection_t *, int);

static LmSSLResponse
jtalk_tls_function(LmSSL *tls, LmSSLStatus status, gpointer dummy)
{
    (void)tls;
    (void)dummy;
    switch(status) {
	case LM_SSL_STATUS_NO_CERT_FOUND:
	    return(LM_SSL_RESPONSE_CONTINUE); /* TODO */
	case LM_SSL_STATUS_UNTRUSTED_CERT:
	    return(LM_SSL_RESPONSE_CONTINUE); /* TODO */
	case LM_SSL_STATUS_GENERIC_ERROR:
	    printf("LM_SSL_STATUS_GENERIC_ERROR\n");
	    return(LM_SSL_RESPONSE_STOP);
	default:
	    printf("TLS status: %d\n", status);
    }
    exit(1);
}

static void
lma_set_available(jtalk_connection_t *con)
{
    LmMessage *m;

    m = lm_message_new_with_sub_type(NULL, LM_MESSAGE_TYPE_PRESENCE,
	LM_MESSAGE_SUB_TYPE_AVAILABLE);
    lm_connection_send(con->con, m, NULL);
    lm_message_unref(m);
}

static void
handle_message(jtalk_connection_t *con, LmMessage *message)
{
    LmMessageType type;
    LmMessageSubType subtype;
    LmMessageNode *node;

    type = lm_message_get_type(message);
    subtype = lm_message_get_sub_type(message);
    node = lm_message_get_node(message);
    if(type == LM_MESSAGE_TYPE_MESSAGE && subtype == LM_MESSAGE_SUB_TYPE_CHAT) {
	const gchar *from;
	LmMessageNode *body;
	const gchar *tbody;

	from = lm_message_node_get_attribute(node, "from");
	body = lm_message_node_get_child(node, "body");
	if(from != NULL && body != NULL)
	    if((tbody = lm_message_node_get_value(body)) != NULL)
		 jtalk_got_chat_text(con, from, tbody);

    } else {
#if 0
	char *s;
	printf("Message: %d / %d.\n", type, subtype);
	node = lm_message_get_node(message);
	s = lm_message_node_to_string(node);
	printf("%s\n", s);
	g_free(s);
#endif
    }
}

static LmHandlerResult
callback_message(LmMessageHandler *handler, LmConnection *connection,
    LmMessage *message, gpointer dummy)
{
    (void)handler;
    (void)connection;

    handle_message(dummy, message);
    return(LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS);
}

static void
callback_authenticated(LmConnection *connection, gboolean success,
    gpointer con_v)
{
    (void)connection;
    jtalk_connection_t *con = con_v;
    GSList *mp;
    LmMessage *m;

    if(!success) {
	printf("Failed to authenticate.\n");
	exit(1);
    }
    jtalk_tell_user("connected as %s to %s.", con->user, con->server);
    con->connected = 1;
    lma_set_available(con);
    con->pending_messages = g_slist_reverse(con->pending_messages);
    for(mp = con->pending_messages; mp != NULL; mp = mp->next) {
	m = mp->data;
	lm_connection_send(con->con, m, NULL);
	lm_message_unref(m);
    }
    g_slist_free(con->pending_messages);
    con->pending_messages = NULL;
}

static void
callback_connected(LmConnection *connection, gboolean success, gpointer con_v)
{
    (void)connection;
    jtalk_connection_t *con = con_v;
    GError *e = NULL;

    if(!success) {
	if(con->reconnect_delay < 500)
	    jtalk_tell_user("failed to connect to %s, retrying.", con->server);
	con->reconnect_delay *= 2;
	if(con->reconnect_delay > 30000)
	    con->reconnect_delay = 30000;
	jtalk_connection_schedule_connect(con, con->reconnect_delay);
	return;
    }
    con->reconnect_delay = 250;
    if(!lm_connection_authenticate(con->con,
	    con->user, con->pass, con->resource,
	    callback_authenticated, con, NULL, &e)) {
	printf("failed to authenticate\n");
	exit(1);
    }
}

static void
callback_disconnected(LmConnection *connection, LmDisconnectReason reason,
    gpointer con_v)
{
    (void)connection;
    jtalk_connection_t *con = con_v;

    con->connected = 0;
    jtalk_tell_user("disconnected from %s (%d), reconnecting...",
	con->server, reason);
    jtalk_connection_schedule_connect(con, 0);
}

static gboolean
jtalk_connection_do_connect(gpointer con_v)
{
    jtalk_connection_t *con = con_v;
    GError *e = NULL;

    if(!lm_connection_open(con->con, callback_connected, con, NULL, &e)) {
	printf("failed to connect\n");
	exit(1);
    }
    return(FALSE);
}

static void
jtalk_connection_schedule_connect(jtalk_connection_t *con, int time)
{
    g_timeout_add(time, jtalk_connection_do_connect, con);
}

jtalk_connection_t *
jtalk_connection_new(char **opts, int nopts)
{
    char *user = NULL;
    char *pass = NULL;
    char *server = NULL;
    char *resource = "jtalk";
    char *jid = NULL;
    int port = 0;
    int use_tls = 0;

    while(nopts > 0) {
	if(strncmp(*opts, "server=", 7) == 0) {
	    server = *opts + 7;
	} else if(strncmp(*opts, "user=", 5) == 0) {
	    user = *opts + 5;
	} else if(strncmp(*opts, "pass=", 5) == 0) {
	    pass = *opts + 5;
	} else if(strncmp(*opts, "resource=", 9) == 0) {
	    pass = *opts + 9;
	} else if(strncmp(*opts, "jid=", 4) == 0) {
	    jid = *opts + 4;
	} else if(strcmp(*opts, "tls") == 0) {
	    use_tls = 1;
	} else if(strncmp(*opts, "port=", 5) == 0) {
	    port = strtol(*opts + 5, NULL, 0);
	} else {
	    fprintf(stderr, "%s: what?\n", *opts);
	    return(NULL);
	}
	opts++;
	nopts--;
    }
    if(server == NULL) {
	fprintf(stderr, "no server\n");
	return(NULL);
    }
    if(user == NULL) {
	fprintf(stderr, "no user name\n");
	return(NULL);
    }
    if(pass == NULL) {
	fprintf(stderr, "no password\n");
	return(NULL);
    }
    
    jtalk_connection_t *con;

    con = g_new(jtalk_connection_t, 1);
    con->server = g_strdup(server);
    con->user = g_strdup(user);
    con->pass = g_strdup(pass);
    con->resource = g_strdup(resource);
    con->jid = jid == NULL ? NULL : g_strdup(jid);
    con->connected = 0;
    con->pending_messages = NULL;

    con->con = lm_connection_new(con->server);
    if(use_tls) {
	con->tls = lm_ssl_new(NULL, jtalk_tls_function, NULL, NULL);
	lm_connection_set_ssl(con->con, con->tls);
    } else {
	con->tls = NULL;
    }
    if(port == 0)
	port = use_tls ? LM_CONNECTION_DEFAULT_PORT_SSL :
	    LM_CONNECTION_DEFAULT_PORT;
    con->port = port;
    lm_connection_set_port(con->con, port);
    if(con->jid != NULL)
	lm_connection_set_jid(con->con, con->jid);
    con->reconnect_delay = 250;

    con->handler = lm_message_handler_new(callback_message, con, NULL);
    lm_connection_register_message_handler(con->con, con->handler,
	LM_MESSAGE_TYPE_MESSAGE, LM_HANDLER_PRIORITY_NORMAL);
    lm_connection_register_message_handler(con->con, con->handler,
	LM_MESSAGE_TYPE_PRESENCE, LM_HANDLER_PRIORITY_NORMAL);
    lm_connection_register_message_handler(con->con, con->handler,
	LM_MESSAGE_TYPE_IQ, LM_HANDLER_PRIORITY_NORMAL);
    lm_connection_register_message_handler(con->con, con->handler,
	LM_MESSAGE_TYPE_STREAM_ERROR, LM_HANDLER_PRIORITY_NORMAL);
    lm_connection_set_disconnect_function(con->con, callback_disconnected,
	con, NULL);

    jtalk_connection_schedule_connect(con, 0);
    return(con);
}

void
jtalk_connection_send_chat_message(jtalk_connection_t *con, jtalk_contact_t *to,
    char *txt)
{
    LmMessage *msg;
    LmMessageNode *node;

    msg = lm_message_new_with_sub_type(to->jid, LM_MESSAGE_TYPE_MESSAGE,
	LM_MESSAGE_SUB_TYPE_CHAT);
    node = lm_message_get_node(msg);
    lm_message_node_add_child(node, "body", txt);
    if(con->connected) {
	lm_connection_send(con->con, msg, NULL);
	lm_message_unref(msg);
    } else {
	con->pending_messages = g_slist_prepend(con->pending_messages, msg);
    }
}
